# Copyright 2023 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Optional, Tuple, Union

import torch

from diffusers import DDPMPipeline, ImagePipelineOutput
from diffusers.utils.torch_utils import randn_tensor

def convert_pipeline(pipeline):
    r"""
    Simple function that converts a huggingface DDPMPipeline into our pipeline that supports step function calls.
    """
    return StepDDPMPipeline(pipeline.unet, pipeline.scheduler)


class StepDDPMPipeline(DDPMPipeline):
    r"""
    Pipeline for image generation.

    This model inherits from [`DiffusionPipeline`]. Check the superclass documentation for the generic methods
    implemented for all pipelines (downloading, saving, running on a particular device, etc.).

    Parameters:
        unet ([`UNet2DModel`]):
            A `UNet2DModel` to denoise the encoded image latents.
        scheduler ([`SchedulerMixin`]):
            A scheduler to be used in combination with `unet` to denoise the encoded image. Can be one of
            [`DDPMScheduler`], or [`DDIMScheduler`].
    """

    model_cpu_offload_seq = "unet"

    def __init__(self, unet, scheduler):
        super().__init__(unet, scheduler)
        
    def denoise_image(self, image, t):
        model_output = self.unet(image, t).sample
        alpha_t = self.scheduler.alphas_cumprod[t]
        image_denoised = (image - (1 - alpha_t).sqrt() * model_output) / alpha_t
        return image_denoised
        
    def get_step_fn(self, generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,):
        def step_fn(image, t):
            # 1. predict noise model_output
            model_output = self.unet(image, t).sample
            
            self.scheduler.alphas_cumprod = self.scheduler.alphas_cumprod.to(image.device)
            # 2. compute previous image: x_t -> x_t-1
            if t.shape and len(t) > 1:
                fn = lambda s, t, x: self.scheduler.step(s[None], t[None], x[None], generator=generator).prev_sample
                # image = torch.vmap(fn)(model_output, t, image)
                out_image = []
                for i in range(len(t)):
                    out_image.append(fn(model_output[i:i+1], t[i:i+1], image[i:i+1]))
                    
                image = torch.cat(out_image, axis=0)
            else:
                image = self.scheduler.step(model_output, t, image, generator=generator).prev_sample

            return image
          
        return step_fn
    
    def eps(self, t, zt):
        t, zt = t.half(), zt.half()
            
        noise_pred = self.unet(zt, t).sample
        
        return noise_pred

    @torch.no_grad()
    def __call__(
        self,
        batch_size: int = 1,
        generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
        num_inference_steps: int = 1000,
        output_type: Optional[str] = "pil",
        return_dict: bool = True,
    ) -> Union[ImagePipelineOutput, Tuple]:
        r"""
        The call function to the pipeline for generation.

        Args:
            batch_size (`int`, *optional*, defaults to 1):
                The number of images to generate.
            generator (`torch.Generator`, *optional*):
                A [`torch.Generator`](https://pytorch.org/docs/stable/generated/torch.Generator.html) to make
                generation deterministic.
            num_inference_steps (`int`, *optional*, defaults to 1000):
                The number of denoising steps. More denoising steps usually lead to a higher quality image at the
                expense of slower inference.
            output_type (`str`, *optional*, defaults to `"pil"`):
                The output format of the generated image. Choose between `PIL.Image` or `np.array`.
            return_dict (`bool`, *optional*, defaults to `True`):
                Whether or not to return a [`~pipelines.ImagePipelineOutput`] instead of a plain tuple.

        Example:

        ```py
        >>> from diffusers import DDPMPipeline

        >>> # load model and scheduler
        >>> pipe = DDPMPipeline.from_pretrained("google/ddpm-cat-256")

        >>> # run pipeline in inference (sample random noise and denoise)
        >>> image = pipe().images[0]

        >>> # save image
        >>> image.save("ddpm_generated_image.png")
        ```

        Returns:
            [`~pipelines.ImagePipelineOutput`] or `tuple`:
                If `return_dict` is `True`, [`~pipelines.ImagePipelineOutput`] is returned, otherwise a `tuple` is
                returned where the first element is a list with the generated images
        """
        # Sample gaussian noise to begin loop
        if isinstance(self.unet.config.sample_size, int):
            image_shape = (
                batch_size,
                self.unet.config.in_channels,
                self.unet.config.sample_size,
                self.unet.config.sample_size,
            )
        else:
            image_shape = (batch_size, self.unet.config.in_channels, *self.unet.config.sample_size)

        if self.device.type == "mps":
            # randn does not work reproducibly on mps
            image = randn_tensor(image_shape, generator=generator)
            image = image.to(self.device)
        else:
            image = randn_tensor(image_shape, generator=generator, device=self.device)

        # set step values
        self.scheduler.set_timesteps(num_inference_steps)
        
        step_fn = self.get_step_fn(generator=generator)

        for t in self.progress_bar(self.scheduler.timesteps):
            image = step_fn(image, t)

        image = (image / 2 + 0.5).clamp(0, 1)
        image = image.cpu().permute(0, 2, 3, 1).numpy()
        if output_type == "pil":
            image = self.numpy_to_pil(image)

        if not return_dict:
            return (image,)

        return ImagePipelineOutput(images=image)